## BLADE core parser
## Auth: H McMILLAN
## See LICENSE file
## (c) Copyright Productivity Commission 2020


blade_variables <- function(..., core_dir = "V:/BLADE1617_CORE/"){
  
  # Load data.table
  if(!("data.table" %in% (.packages()))) library(data.table)
  
  # Header extraction function
  header <- function(dataset){
    paths <- list.files(
      path        = core_dir, 
      pattern     = paste0(dataset, ".*CSV"),   
      full.names  = T, 
      recursive   = T, 
      ignore.case = T
      )
    headers <- unlist(lapply(paths, function(p) names(fread(p, nrows = 0))))
    data.table(name = unique(headers), data = dataset)
  }
  
  # Collect headers
  rbindlist(
    list(
      header("FRAME"),
      header("BIT"),
      header("BAS"),
      header("PAYG")
    )
  )
}



parse_blade <- function(select, outer_join_frame = FALSE, ...){
  
  # Initial checks
  stopifnot(is.character(select) | is.null(select))
  stopifnot(is.logical(outer_join_frame))
  
  # Start timer
  start_time <- proc.time()
  
  # Error handler
  "%iferror%" <- function(a, b) tryCatch({a}, error = function(e) {b})
  
  # Error handling get
  safe_get <- function(x, ...) get(x, ...) %iferror% {
    warning(paste0("Error: cannot find ", deparse(substitute(x))))
    NULL
  }
  
  # Get, merge and remove data from parent envir with garbage collect
  # Only works on data.tables with keys to prevent copy on modify + speed
  get_merge_rm <- function(a, b){
    if(is.character(a)){df1 <- safe_get(a)} else {df1 <- a}
    if(is.character(b)){df2 <- safe_get(b)} else {df2 <- b}
    
    df <- merge.data.table(df1, df2, all = T)
    
    if(is.character(a)){
      rm(list = c(a, "df1"), inherits = T) %iferror% NULL
    } else { 
      rm(a, df1, inherits = T) %iferror% NULL
    }
    
    if(is.character(b)){
      rm(list = c(b, "df2"), inherits = T) %iferror% NULL
    } else { 
      rm(b, df2, inherits = T) %iferror% NULL
    }
    
    gc()
    return(df)
  }
  
  # Get list of avaialable variables by dataset
  vars <- blade_variables()
  
  # Check for NULL select (parse all)
  if(!is.null(select)){
    
    # Add names list without BIT type of organisation prefix
    vars[, sans_bit_pre := gsub("^C_|^I_|^P_|^T_", "", name, ignore.case = T)]
    
    # Remove mandatory merging variables from select
    select <- select[grep("^id|^date", select, ignore.case = T, invert = T)]
    
    # Check for select values without matches
    no_match <- select[
      sapply(
        select, 
        function(sel){
          sel <- paste0("^", sel)
          !any(grepl(sel, c(vars$name, vars$sans_bit_pre), ignore.case = T))
        } 
      )
      ]
    
    # Remove values without matches
    select <- select[!(select %in% no_match)]
    
    # Loop through datasets and drop non-matches in vars
    # Used a loop to prevent simple variable names (e.g. div) from
    # matching to longer names in datasets like BIT - hense the 
    # dataset loop order.
    for(dset in c("FRAME", "PAYG", "BAS", "BIT")){
      # Find matching vars as recorded in vars dataset
      matched_vars <- lapply(
        select, 
        function(sel){
          vars[data == dset, c(name, sans_bit_pre)][
            grepl(
              sel, 
              vars[data == dset, c(name, sans_bit_pre)], 
              ignore.case = T
            )
          ]
        } 
      )
      # Find matching vars as recorded in select vector
      matched_select <- select[
        sapply(
          select, 
          function(sel){
            any(grepl(
              sel, 
              vars[data == dset, c(name, sans_bit_pre)], 
              ignore.case = T
            ))
          } 
        )
      ]
      # Only keep uniqe matches (duplicates due to sans_bit_pre)
      matched_vars <- unique(unlist(matched_vars))
      # Drop non-match from vars
      vars[data == dset & !(name %in% matched_vars), drop := T]
      vars <- vars[is.na(drop)]
      # Drop matched variables from select before next loop
      select <- select[!(select %in% matched_select)]
    }
    
    #Remove drop and sans_bit_pre column from vars
    vars[, c("drop", "sans_bit_pre") := NULL]
    
    # Message parse summary
    message("PARSE SUMMARY:\n")
    message(paste0(capture.output(vars), collapse = "\n"), "\n")
    if(length(no_match > 0)){
      warning(
        paste0(no_match, collapse = ", "), 
        " not available in BLADE core CSV files"
      )
    } 
  }
  
  # Parse FRAME
  if("FRAME" %in% vars$data){
    # Check if select is NULL
    frame_select <- if(is.null(select)){
      select
    } else {
      vars[data == "FRAME", c("id", name)]
    }
    # Call parser
    frame <- parse_frame(select = frame_select, ...)
  }
  
  # Parse BIT
  if("BIT" %in% vars$data){
    # Check if select is NULL
    bit_select <- if(is.null(select)){
      select
    } else {
      vars[data == "BIT", c("id", name)]
    }
    # Call parser
    bit <- parse_bit(select = bit_select, ...)
  }
  
  # Parse BAS
  if("BAS" %in% vars$data){
    # Check if select is NULL
    bas_select <- if(is.null(select)){
      select
    } else {
      vars[data == "BAS", c("id", name)]
    }
    # Call parser
    bas <- parse_bas(select = bas_select, ...)
  }
  
  # Parse PAYG
  if("PAYG" %in% vars$data){
    # Check if select is NULL
    payg_select <- if(is.null(select)){
      select
    } else {
      vars[data == "PAYG", c("id", name)]
    }
    # Call parser
    payg <- parse_payg(select = payg_select, ...)
  }
  
  # Harmonise keys
  lapply(
    unique(vars$data),
    function(dt){
      message("Setting ", dt, " keys")
      setkey(get(tolower(dt)), date, id)
    }
  )
  
  # Merge
  message("Merging BLADE core")
  to_merge <- tolower(unique(vars$data))
  
  # Case 1 inner join frame
  if(
    "frame" %in% to_merge & 
    outer_join_frame == F &
    length(to_merge) > 1
    ){
    not_frame <- to_merge[to_merge != "frame"]
    blade     <- if(length(not_frame) > 1){
      Reduce(get_merge_rm, not_frame)
    } else {
      safe_get(not_frame[1])
    }
    blade <- frame[blade]
    rm(frame, inherits = T) %iferror% NULL
    
  # Case 2 outer join 
  } else {
    blade <- if(length(to_merge) > 1){
      Reduce(get_merge_rm, to_merge)
    } else {
      safe_get(to_merge)
    }
  } 
  
  ## Stop timer
  end_time <- proc.time()
  
  ## Message duration
  duration_time <- round((end_time - start_time)[3])
  duration_time <- lubridate::seconds_to_period(duration_time)
  duration_time <- paste("Time to parse BLADE:", duration_time)
  message(duration_time)
  
  # Return
  return(blade)

}




