## Frame Parser
## Auth: H McMILLAN
## See LICENSE file
## (c) Copyright Productivity Commission 2020

parse_frame <- function(
  select      = NULL,
  clean_vars  = TRUE,
  clean_id    = clean_vars,
  clean_bday  = clean_vars,
  clean_div   = clean_vars,
  clean_state = clean_vars,
  clean_sisca = clean_vars,
  clean_al_st = clean_vars,
  paths       = 
    c(
      "V:/BLADE1617_CORE/CSV/IN22_BLADE1617_FRAME_0102.csv",
      "V:/BLADE1617_CORE/CSV/IN22_BLADE1617_FRAME_0203.csv",
      "V:/BLADE1617_CORE/CSV/IN22_BLADE1617_FRAME_0304.csv",
      "V:/BLADE1617_CORE/CSV/IN22_BLADE1617_FRAME_0405.csv",
      "V:/BLADE1617_CORE/CSV/IN22_BLADE1617_FRAME_0506.csv",
      "V:/BLADE1617_CORE/CSV/IN22_BLADE1617_FRAME_0607.csv",
      "V:/BLADE1617_CORE/CSV/IN22_BLADE1617_FRAME_0708.csv",
      "V:/BLADE1617_CORE/CSV/IN22_BLADE1617_FRAME_0809.csv",
      "V:/BLADE1617_CORE/CSV/IN22_BLADE1617_FRAME_0910.csv",
      "V:/BLADE1617_CORE/CSV/IN22_BLADE1617_FRAME_1011.csv",
      "V:/BLADE1617_CORE/CSV/IN22_BLADE1617_FRAME_1112.csv",
      "V:/BLADE1617_CORE/CSV/IN22_BLADE1617_FRAME_1213.csv",
      "V:/BLADE1617_CORE/CSV/IN22_BLADE1617_FRAME_1314.csv",
      "V:/BLADE1617_CORE/CSV/IN22_BLADE1617_FRAME_1415.csv",
      "V:/BLADE1617_CORE/CSV/IN22_BLADE1617_FRAME_1516.csv",
      "V:/BLADE1617_CORE/CSV/IN22_BLADE1617_FRAME_1617.csv"
    ),
  ...
){
  
  ## Start timer
  start_time <- proc.time()
  
  ## Argument check
  stopifnot(is.character(select) | is.null(select))
  stopifnot(is.character(paths))
  stopifnot(all(file.exists(paths)))
  stopifnot(is.logical(clean_vars))
  stopifnot(is.logical(clean_bday))
  stopifnot(is.logical(clean_div))
  stopifnot(is.logical(clean_state))
  
  ## Check for tsid (for date parsing)
  select_cols  <- !is.null(select)
  require_tsid <- select_cols & "tsid" %in% select
  
  if(select_cols & !require_tsid){
    select <- c(select, "tsid")
  }
  
  ## Required package
  if(!("data.table" %in% (.packages()))) library(data.table)
  
  ## Define fread with messaged file name
  mfread <- function(path, cols = select, ...){
    message("Parsing ", basename(path))
    fread(path, select = cols, stringsAsFactors = T, ...)
  }
  
  ## List files and read in
  frame <- lapply(paths, mfread)
  
  ## Name and rowbind
  message("Binding datasets")
  names(frame) <- basename(tools::file_path_sans_ext(paths))
  frame        <- rbindlist(frame, fill = T)
  
  ## Clean dates
  message("Adding new `date` column (FY end)")
  dates <- seq.Date(
    as.Date("2001-06-30"),
    as.Date("2017-06-30"),
    "years"
  )
  frame[, date := dates[tsid]]
  
  ## Drop tsid if not required
  if(select_cols & !require_tsid){
    frame[, tsid := NULL]
  }
  
  ## Clean IDs
  if("id" %in% names(frame) & clean_id){
    message("Reclassing `id` to integer")
    frame[, id := as.integer(stringr::str_sub(id, 3, -1))]
  }
  
  ## Clean BG IDs
  if("bg_id" %in% names(frame) & clean_id){
    message("Reclassing `bg_id` to integer")
    frame[, bg_id := as.integer(stringr::str_sub(bg_id, 3, -1))]
  }
  
  ## Clean birthday
  if("birth_date" %in% names(frame) & clean_bday){
    message("Reclassing birthday to date (FY end)")
    frame[, birth_date := stringr::str_sub(birth_date, 1, 4)]
    frame[, birth_date := factor(birth_date)]
    bdays <- lubridate::ymd(
      paste0(
        as.numeric(levels(frame$birth_date)) + 1, 
        "-06-30"
      ),
      quiet = TRUE
    )
    frame[, birth_date := bdays[birth_date]]
  }
  
  ## Clean industry division
  if("Div" %in% names(frame) & clean_div){
    message("Cleaning industry division labels")
    industries <- factor(
      c(
        "Agriculture, Forestry and Fishing",
        "Mining",
        "Manufacturing",
        "Electricity, Water and Waste Services",
        "Construction",
        "Wholesale Trade",
        "Retail Trade",
        "Accommodation and Food Services",
        "Transport, Postal and Warehousing",
        "Information Media and Telecommunications",
        "Finance and Insurance Services",
        "Rental, Hiring and Real Estate Services",
        "Professional, Scientific and Technical Services",
        "Administrative and Support Services",
        "Public Administration and Safety",
        "Education and Training",
        "Healthcare and Social Assistance",
        "Arts and Recreation Services",
        "Other Services",
        "Unknown or missing"
      )
    )
    frame[, Div := industries[match(Div, LETTERS)]]
  }
  
  ## Clean states
  if("x_state" %in% names(frame) & clean_state){
    message("Cleaning state/territory labels")
    states <- factor(
      c(
        NA,
        "NSW",
        "VIC",
        "QLD",
        "SA",
        "WA",
        "TAS",
        "NT",
        "ACT"
      )
    )
    frame[, x_state := states[x_state + 1]]
  }
  
  ## Clean Sisca
  if("X_SISCA08" %in% names(frame) & clean_sisca){
    message("Cleaning SISCA labels") 
    sisca <- data.table(
      X_SISCA08 = c(
        0, 1001, 1009, 2110, 2121, 2129, 2131, 2132, 2133,
        2141, 2142, 2191, 2199, 2200, 2301, 2309, 3000,
        4000, 5000, 6000
      ),
      sisca = factor(
        c(
          "Unknown or missing",
          "Non-Financial Investment Funds",
          "Other Non-Financial Corporations",
          "Reserve Bank of Australia",
          "Banks",
          "Other Depository Corporations",
          "Superannuation Funds",
          "Life Insurance Corporations",
          "Non-Life Insurance Corporations",
          "Money Market Funds",
          "Non-Money Market Financial Investment Funds",
          "Securities",
          "Other Financial Intermediaries",
          "Financial Auxiliaries",
          "Central Borrowing Authorities",
          "Money Lenders and Other Captive Financial Institutions",
          "General Government",
          "Households",
          "Non-profit Institutions Serving Households",
          "Rest of the World"
        )
      )
    )
    
    frame <- sisca[frame, on = "X_SISCA08"]
    frame[, X_SISCA08 := NULL]
    setnames(frame, "sisca", "X_SISCA08")
  }
  
  # Clean alive status
  if("x_al_st" %in% names(frame) & clean_al_st){
    message("Cleaning alive status labels") 
    alive <- data.table(
      x_al_st = c(0L, 1L, 2L, 7L, 9L),
      alive   = factor(
        c(
          "Dead / Cancelled",
          "Alive / Active",
          "Excluded units: not enough information available to stratify",
          "Suppressed ABN",
          "ATO Client / ABN linked to profiled population"
        )
      )
    )
    frame <- alive[frame, on = "x_al_st"]
    frame[, x_al_st := alive]
    frame[, alive   := NULL]
  }
  
  ## Stop timer
  end_time <- proc.time()
  
  ## Message duration
  duration_time <- round((end_time - start_time)[3])
  duration_time <- lubridate::seconds_to_period(duration_time)
  duration_time <- paste("Time to parse FRAME:", duration_time)
  message(duration_time)
  
  ## Return bound data
  return(frame)
  
}