## Train xgboost

## Auth: H McMILLAN
## See LICENSE file
## (c) Copyright Productivity Commission 2020


## Packages
library(data.table)
library(magrittr)
library(pbapply)
library(xgboost)
library(caret)
library(fst)



## Options 
reeval_existing_models <- F



## Model params
param <- list(
  eta       = 0.3,
  gamma     = 0,
  max_depth = 6
)



## Model list 
impute_list <- list.files(
  path       = "./xgboost data/",
  pattern    = "matrix_.+_test.fst"
)
impute_list <- gsub(
  pattern     = "matrix_|_test.fst",
  replacement = "",
  x           = impute_list
)




## Find already completed models
complete_list <- list.files(
  path       = "./xgboost data/",
  pattern    = "xgb_model_.+.rds"
)
complete_list <- gsub(
  pattern     = "xgb_model_|.rds",
  replacement = "",
  x           = complete_list
)

if(reeval_existing_models == F){
  impute_list <- impute_list[!(impute_list %in% complete_list)]
}





## Model creation function
make_xgb <- function(var, verbose = TRUE){
  # Message if verbose function
  v_message <- function(...) if(verbose == T) message(...) 
  
  # Define matrix paths
  train_path <- paste0("./xgboost data/matrix_", var, "_train.fst")
  test_path  <- paste0("./xgboost data/matrix_", var, "_test.fst")
  
  # Read in matricies as xgboost matrix
  v_message("Parsing and reclassing train matrix")
  dtrain <- as.matrix(read_fst(train_path)) 
  dtrain <- xgb.DMatrix(
      data  = dtrain[, which(colnames(dtrain) != var)],
      label = dtrain[, var]
  )
  gc(verbose = F)
  
  v_message("Parsing and reclassing test matrix")
  dtest <- as.matrix(read_fst(test_path)) 
  dtest <- xgb.DMatrix(
      data  = dtest[, which(colnames(dtest) != var)],
      label = dtest[, var]
    )
  gc(verbose = F)
  
  # Model
  v_message("Training model")
  xgb <- xgb.train(
    params    = param,
    nrounds   = 20,
    seed      = 88,
    data      = dtrain,
    watchlist = list(test  = dtest, train = dtrain)
  )
  
  # Save results
  v_message("Saving results")
  list(
    model      = xgb, 
    test_y     = getinfo(dtest, "label"),
    test_y_hat = predict(xgb, dtest)
    ) %>% 
    saveRDS(paste0("./xgboost data/xgb_model_", var, ".rds"))
  
  # Make sure R is flushing the environment
  v_message("Releasing memory allocations")
  rm(dtest, dtrain, xgb)
  gc(verbose = F)
  
  # return
  v_message(var, " complete")
  return()
  
}




## eval

sapply(impute_list, make_xgb)


