## Impute missing values

## Auth: H McMILLAN
## See LICENSE file
## (c) Copyright Productivity Commission 2020+


## Packages
library(data.table)
library(magrittr)
library(pbapply)
library(xgboost)
library(caret)
library(fst)




## Options 
## (!) Don't skip this or xgboost will only run on complete cases
options(na.action = "na.pass")
options(scipen = 9999L)



## Get utility functions
list.files("./R/", ".r|.R", F, T) %>% 
  sapply(source)




## Define required variables from BLADE core 
## BLADE parser accepts regex name matches
variables <- c(
  "id", "div", "x_anzsic06", "^date",
  "fte", "turnover", "wages", "birth_date", "costsale", 
  "TOTLEXPS",  "x_al_st",
  "capex", "oexp", "credit_for_gst_paid",  
  "GST_PAYABLE",  
  "P_WRK_AMT",  "I_INCM_AMT", "hcnt", 
  "TOTLWAGE", "TERINDEP", "TEROTHDE",
  "COSTSALE", "DEPREXPS", "CLOSTOCK", "OPNSTOCK", 
  "OTHDEPAS", "TERINTAS", "X_SECTOR", "X_TOLO", "X_STATE", 
  "INTDEPAS", "LABEL_D"
)




## Check if you have a copy of BLADE with required variables
`got BLADE?` <- {
  if(file.exists("data/blade.fst")){
    blade_names <- names(read_fst("data/blade.fst", to = 1))
    verif_names <- grep(paste(variables, collapse = "|"), blade_names, T, F, T)
    all(sapply(variables, function(x) any(grepl(x, blade_names, T))))
  } else {
    FALSE
  }
}




## If you got it, load it
if(`got BLADE?`) blade <- read_fst(
  path          = "data/blade.fst", 
  columns       = verif_names,
  as.data.table = T
)




## if not, construct a BLADE panle dataset from csv files
## This will take 1 to 2 hours depending on Datalab server
if(!`got BLADE?`) {
  blade <- parse_blade(
    select           = variables,
    outer_join_frame = F,
    collapse         = T,
    collapse_fun     = "sum"
  )
  write.fst(blade, "data/blade.fst")
}




## Find column duplications due to inconsistent character cases in BIT
badnames <- names(blade) %>% 
  tolower %>% 
  duplicated %>% 
  names(blade)[.] %>% 
  lapply(function(x) grep(x, names(blade), T, F, T))




## Take the max of duplicated columns
sapply(badnames, function(x){
  blade[, (x[1]) := do.call(pmax, c(.SD, na.rm = T)), .SDcols = x]
  blade[, (x[2:length(x)]) := NULL]
  return()
})




## Make wages max of BAS and BIT wage cols
blade[, WAGES    := pmax(WAGES, TOTLWAGE, na.rm = T)]
blade[, TOTLWAGE := NULL]




## Change FTE to FTE + 1
blade[, fte := fte + 1]




## Filter
blade <- blade[
  (TURNOVER >= 0 | is.na(TURNOVER)) &
  (WAGES    >= 0 | is.na(WAGES))    &
  (fte      >= 0 | is.na(fte))
]




## Lowercase all column names
names(blade) %<>% tolower




## Define log transformation function for numerical variables
## As per Suresh, Taib, Zhao and Jin (2019)
tform   <- function(num) log10(num + 1e-6)




## Identify numeric valiables 
## & exclude from discrete categoricals classed as integers
numeric_vars <- names(blade)[sapply(blade, is.numeric)]
numeric_vars <- numeric_vars[
  !(numeric_vars %in% c("id", "x_pcode", "x_sector", "x_anzsic06", "x_tolo"))
  ]




## Create log versions of numeric variables
blade[
  ,
  (paste0("log_", numeric_vars)) := lapply(.SD, tform),
  .SDcols = numeric_vars
  ]




## Make sure categorical variables are factors
to_factor <- c("x_anzsic06", "x_tolo", "x_state", "x_sector")
blade[, (to_factor) := lapply(.SD, factor), .SDcols = to_factor]




model_formulas <- list(
  
  fte = 
    ~ log_fte + 
    x_tolo + x_sector + x_state + div + birth_date + date + 
    log_totlexps + log_intdepas + log_othdepas + log_terindep + 
    log_terothde + log_costsale + log_opnstock + log_clostock + log_deprexps + 
    log_terintas + log_turnover + log_capex + log_oexp + 
    log_credit_for_gst_paid + log_gst_payable + log_wages + 
    log_p_wrk_amt + log_i_incm_amt + log_label_d - 1,
  
  turnover = 
    ~ log_turnover + 
    x_tolo + x_sector + x_state + div + birth_date + date + 
    log_totlexps + log_intdepas + log_othdepas + log_terindep + 
    log_terothde + log_costsale + log_opnstock + log_clostock + log_deprexps + 
    log_terintas + log_capex + log_oexp + log_credit_for_gst_paid + 
    log_gst_payable + log_wages + log_p_wrk_amt + log_i_incm_amt + log_fte +
    log_hcnt + log_label_d - 1,
  
  opnstock = 
    ~ log_opnstock +
    x_tolo + x_sector + x_state + div + birth_date + date + 
    log_totlexps + log_intdepas + log_othdepas + log_terindep + 
    log_terothde + log_costsale + log_deprexps + log_terintas + 
    log_turnover + log_capex + log_oexp + log_credit_for_gst_paid + 
    log_gst_payable + log_wages + log_p_wrk_amt + log_i_incm_amt + log_fte +
    log_hcnt + log_label_d - 1, 
  
  clostock = 
    ~ log_clostock +
    x_tolo + x_sector + x_state + div + birth_date + date + 
    log_totlexps + log_intdepas + log_othdepas + log_terindep + 
    log_terothde + log_costsale + log_deprexps + log_terintas + 
    log_turnover + log_capex + log_oexp + log_credit_for_gst_paid + 
    log_gst_payable + log_wages + log_p_wrk_amt + log_i_incm_amt + log_fte +
    log_hcnt + log_label_d - 1, 
  
  capex = 
    ~ log_capex + 
    x_tolo + x_sector + x_state + div + birth_date + date + 
    log_totlexps + log_intdepas + log_othdepas + log_terindep + 
    log_terothde + log_costsale + log_opnstock + log_clostock + log_deprexps + 
    log_terintas + log_turnover + log_oexp + log_credit_for_gst_paid + 
    log_gst_payable + log_wages + log_p_wrk_amt + log_i_incm_amt + log_fte + 
    log_hcnt + log_label_d - 1, 
  
  oexp = 
    ~ log_oexp +
    x_tolo + x_sector + x_state + div + birth_date + date + 
    log_totlexps + log_intdepas + log_othdepas + log_terindep + 
    log_terothde + log_costsale + log_opnstock + log_clostock + log_deprexps + 
    log_terintas + log_turnover + log_capex +  log_credit_for_gst_paid + 
    log_gst_payable + log_wages + log_p_wrk_amt + log_i_incm_amt + log_fte + 
    log_hcnt + log_label_d - 1,
  
  terothde = 
    ~ log_terothde + 
    x_tolo + x_sector + x_state + div + birth_date + date + 
    log_totlexps + log_intdepas + log_othdepas + log_terindep + 
    log_costsale + log_opnstock + log_clostock + log_deprexps + log_terintas + 
    log_turnover + log_capex + log_oexp + log_credit_for_gst_paid + 
    log_gst_payable + log_wages + log_p_wrk_amt + log_i_incm_amt + log_fte + 
    log_hcnt + log_label_d - 1,
  
  label_d = 
    ~ log_label_d +
    x_tolo + x_sector + x_state + div + birth_date + date + 
    log_totlexps + log_intdepas + log_othdepas + log_terindep + 
    log_terothde + log_costsale + log_opnstock + log_clostock + log_deprexps + 
    log_terintas + log_turnover + log_capex + log_oexp + 
    log_credit_for_gst_paid + log_gst_payable + log_wages + 
    log_p_wrk_amt + log_i_incm_amt + log_fte + log_hcnt - 1
  
)





## Create imputation function
impute_missing <- function(model_formula){
  
  # Get y var
  label_var <- deparse(model_formula) %>% 
    strsplit("~| |\\+") %>% 
    unlist() %>% 
    .[. != ""] %>% 
    .[1]
  
  # Create matrix
  message("Creating matrix")
  m <- blade[is.na(get(label_var))] %>% 
    model.matrix(model_formula, .) 
  
  # Convert matrix to xgb.DMatrix
  m <- xgb.DMatrix(m[, which(colnames(m) != label_var)])
  gc(verbose = F)
  
  # Read in relevant model
  message("Reading in model")
  mod <- readRDS(paste0("xgboost data/xgb_model_", label_var, ".rds"))$model
  
  # Predict
  message("Predicting missing values")
  blade[is.na(get(label_var)), (label_var) := predict(mod, m)]
  
  # Clean envir
  message("Cleaning environment")
  rm(mod, m)
  gc(verbose = F)
  
  # return
  return()
}


## Run imputation
pblapply(model_formulas, impute_missing)



## Reverse log transformation on imputed values
untform <- function(x) 10 ^ x - 1e-6

pblapply(numeric_vars, function(x){
  log_name <- paste0("log_", x)
  blade[
    , 
    (paste0("imp_flag_", x)) := 
      ifelse(is.na(get(x)) & !is.na(get(log_name)), T, F)
  ]
  blade[
    is.na(get(x)) & !is.na(get(log_name)), 
    (x) := untform(get(log_name))
  ]
  blade[, (log_name) := NULL]
  return()
})




## Save 
write_fst(blade, "data/BLADE fully imputed + flags.fst")




